# Disk resonator custom component definition (layout, netlist and circuitmodel).

from ipkiss3 import all as i3
import numpy as np


class DiskResonator(i3.PCell):
    """A disk resonator: a circular waveguide evanescently coupled to a straight bus waveguide." \
    """

    class Layout(i3.LayoutView):
        disk_radius = i3.PositiveNumberProperty(default=10.0, doc="Disk radius in um")
        disk_wg_width = i3.PositiveNumberProperty(default=0.5, doc="Width of the bus waveguide in um")
        coupler_spacing = i3.PositiveNumberProperty(
            default=0.6,
            doc="Spacing between the centerline of the bus waveguide and the edge of the disk",
        )

        def _generate_elements(self, elems):
            # Define disk and bus shape
            shape_disk = i3.ShapeCircle(center=(0, 0), radius=self.disk_radius)
            shape_bus = [
                (-self.disk_radius, -self.disk_radius - self.coupler_spacing),
                (self.disk_radius, -self.disk_radius - self.coupler_spacing),
            ]

            # Add the shapes to elems
            elems += i3.Boundary(layer=i3.TECH.PPLAYER.SI, shape=shape_disk)
            elems += i3.Path(layer=i3.TECH.PPLAYER.SI, shape=shape_bus, line_width=self.disk_wg_width)

            return elems

        def _generate_ports(self, ports):
            ports += i3.OpticalPort(
                name="in", position=(-self.disk_radius, -self.disk_radius - self.coupler_spacing), angle=180.0
            )
            ports += i3.OpticalPort(
                name="out", position=(self.disk_radius, -self.disk_radius - self.coupler_spacing), angle=0.0
            )
            return ports

    class Netlist(i3.NetlistView):
        def _generate_terms(self, terms):
            terms += i3.OpticalTerm(name="in")
            terms += i3.OpticalTerm(name="out")
            return terms

    class CircuitModel(i3.CircuitModelView):
        # Try changing these default values to see how the similarity behaves.
        disk_n_eff = i3.PositiveNumberProperty(default=3.0, doc="Effective index of the disk waveguide")
        disk_length = i3.PositiveNumberProperty(default=10.0, doc="Disk physical length")
        disk_roundtrip_transmission = i3.NonNegativeNumberProperty(
            default=0.92, doc="Amplitude transmission after one pass in the disk"
        )
        coupler_bar_transmission = i3.ComplexNumberProperty(
            default=0.94,
            doc="Transmission in the (lossless) directional coupler;"
            " what passes through and is not dropped to the other channel.",
        )

        def _generate_model(self):
            return DiskResonatorModel(
                disk_n_eff=self.disk_n_eff,
                disk_length=self.disk_length,
                disk_roundtrip_transmission=self.disk_roundtrip_transmission,
                coupler_bar_transmission=self.coupler_bar_transmission,
            )


class DiskResonatorModel(i3.CompactModel):
    """Behavioral model for a disk resonator"""

    parameters = ["disk_n_eff", "disk_length", "coupler_bar_transmission", "disk_roundtrip_transmission"]
    terms = [i3.OpticalTerm(name="in"), i3.OpticalTerm(name="out")]

    def calculate_smatrix(parameters, env, S):
        beta = 2.0 * np.pi / env.wavelength * parameters.disk_n_eff
        numer = parameters.coupler_bar_transmission - parameters.disk_roundtrip_transmission * np.exp(
            1j * beta * parameters.disk_length
        )
        denom = 1.0 - parameters.coupler_bar_transmission * parameters.disk_roundtrip_transmission * np.exp(
            1j * beta * parameters.disk_length
        )

        S["in", "out"] = S["out", "in"] = numer / denom
